module FEQuad

using LinearAlgebra
using SparseArrays
using Logging

using HCubature

using ...CoordinateSystem
using ..FEMaterial
using ..FENode

export Quad

mutable struct Quad
    id::String
    hid::Int

    node1::Node
    node2::Node
    node3::Node
    node4::Node
    material::Material
    t::Float64

    membrane::Bool
    bending::Bool

    center::Array{Float64}
    A::Float64
    T::SparseMatrixCSC{Float64}

    Kmᵉ::SparseMatrixCSC{Float64}
    Kbᵉ::SparseMatrixCSC{Float64}

    Kᵉ::SparseMatrixCSC{Float64}
    Mᵉ::SparseMatrixCSC{Float64}
end

function Quad(id,hid,node1,node2,node3,node4,material,t,membrane=true,bending=true)
    #Initialize local CSys,could be optimized by using a MSE plane
    o=(node1.loc+node2.loc+node3.loc+node4.loc)/4
    pt1 = (node2.loc+node3.loc)/2
    pt2 = (node3.loc+node4.loc)/2
    csys = CSys(o, pt1, pt2)

    vertices=[n.loc for n in (node1,node2,node3,node4)]
    #check aspect ratio
    a=vertices[1]-vertices[3]
    b=vertices[2]-vertices[4]
    if !(0.1<norm(a)/norm(b)<10)
        @warn "Quad "*string(id)*" has an aspect ratio larger than 10, may lead to element illness."
    end
    #check wrong sequence
    for i in 1:4
        j=i%4+1
        if (vertices[i]-o)×(vertices[j]-o)⋅csys.T[:,3]<0
            @warn  "Quad "*string(id)*" has a wrong sequence of nodes, may lead to element illness."
            break
        end
    end
#    check devolution vertex
    for i in 1:4
        j=i%4+1
        if norm(vertices[i]-vertices[j])<1e-9
            @warn  "Quad "*string(id)*" has only 3 independent nodes, may lead to element illness."
            break
        end
    end
    #check distorsion
    for i in 1:4
        v=vertices[i]-o
        if (v⋅csys.T[:,3]/norm(csys.T[:,3]))/norm(v)>0.1
            @warn  "Quad "*string(id)*" is too distorted, may lead to element illness."
            break
        end
    end
    #check convexness
    for i in 1:4
        j=(i+1)%4+1
        k=i%4+1
        if norm((vertices[i]+vertices[j])/2-o)>=norm(vertices[k]-o)
            @warn  "Quad "*string(id)*" is not convex, may lead to element illness."
            break
        end
    end

    #area is considered as the average of trangles generated by splitting the quand with diagonals
    A=0.5*det([[1;1;1] node1.loc-o node2.loc-o]')
    A+=0.5*det([[1;1;1]  node2.loc-o node3.loc-o]')
    A+=0.5*det([[1;1;1] node3.loc-o node4.loc-o]')
    A+=0.5*det([[1;1;1] node4.loc-o node1.loc-o]')

    #3D to local 2D
    T=zeros(24,24)
    for i in 1:8
        T[3i-2:3i,3i-2:3i]=csys.T
    end
    T=sparse(T)

    Kbᵉ=spzeros(1,1)
    Kmᵉ=spzeros(1,1)

    Kᵉ=spzeros(1,1)
    Mᵉ=spzeros(1,1)

    Quad(id,hid,node1,node2,node3,node4,material,t,membrane,bending,o,A,T,Kbᵉ,Kmᵉ,Kᵉ,Mᵉ)
end

function integrateKm!(elm::Quad)::SparseMatrixCSC{Float64}
    E₀,ν₀=elm.material.E,elm.material.ν
    center=elm.center
    t=elm.t
    T=elm.T[1:3,1:3]
    x₁,y₁,z₁=T*(elm.node1.loc-center)
    x₂,y₂,z₂=T*(elm.node2.loc-center)
    x₃,y₃,z₃=T*(elm.node3.loc-center)
    x₄,y₄,z₄=T*(elm.node4.loc-center)

    K=Matrix{Float64}(undef,12,12)
    J=Matrix{Float64}(undef,2,2)

    function BtDB(x)
        ξ,η=x[1],x[2]
        K[1,1]=0.0625*(1 - η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,2]=0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,3]=-0.03125*(1 - η)*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) + 0.03125*(1 - η)*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[1,4]=-0.0625*(1 - η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,5]=-0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2)
        K[1,6]=-0.03125*(1 - η)*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) + 0.03125*(1 - η)*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[1,7]=-0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,8]=-0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2)
        K[1,9]=-0.03125*(1 - η)*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) + 0.03125*(1 - η)*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[1,10]=0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,11]=-0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[1,12]=-0.03125*(1 - η)*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) + 0.03125*(1 - η)*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[2,1]=0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,2]=0.0625*(1 - ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,3]=0.03125*(1 - ξ)*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) - 0.03125*(1 - ξ)*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[2,4]=-0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,5]=-0.03125*(1 - η)^2*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2)
        K[2,6]=0.03125*(1 - ξ)*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 - η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2) - 0.03125*(1 - ξ)*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[2,7]=-0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,8]=-0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2) - 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,9]=0.03125*(1 - ξ)*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2) - 0.03125*(1 - ξ)*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[2,10]=-0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[2,11]=-0.0625*(1 - ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[2,12]=0.03125*(1 - ξ)*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) - 0.03125*(1 - ξ)*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[3,1]=-0.25*(1 - η)*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,2]=-0.25*(1 - ξ)*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,3]=0.125*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))^2/(1 - ν₀^2)
        K[3,4]=0.25*(1 - η)*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,5]=-0.25*(1 + ξ)*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,6]=-0.125*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + 0.125*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[3,7]=0.25*(1 + η)*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,8]=0.25*(1 + ξ)*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,9]=0.125*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2))*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[3,10]=-0.25*(1 + η)*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,11]=0.25*(1 - ξ)*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[3,12]=0.125*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))*(0.125*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[4,1]=-0.0625*(1 - η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,2]=-0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,3]=0.03125*(1 - η)*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.03125*(1 - η)*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[4,4]=0.0625*(1 - η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 + ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,5]=-0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,6]=0.03125*(1 - η)*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.03125*(1 - η)*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[4,7]=0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 + ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,8]=0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,9]=0.03125*(1 - η)*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.03125*(1 - η)*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[4,10]=-0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,11]=0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[4,12]=0.03125*(1 - η)*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.03125*(1 - η)*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[5,1]=-0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2)
        K[5,2]=-0.03125*(1 - η)^2*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2)
        K[5,3]=0.03125*(1 + ξ)*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) - 0.03125*(1 + ξ)*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[5,4]=-0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[5,5]=0.0625*(1 + ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[5,6]=0.03125*(1 + ξ)*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 - η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2) - 0.03125*(1 + ξ)*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[5,7]=0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[5,8]=-0.0625*(1 + ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[5,9]=0.03125*(1 + ξ)*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2) - 0.03125*(1 + ξ)*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[5,10]=0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[5,11]=-0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2) - 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[5,12]=0.03125*(1 + ξ)*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) - 0.03125*(1 + ξ)*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[6,1]=-0.25*(1 - η)*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,2]=-0.25*(1 - ξ)*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) - 0.125*(1 - η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,3]=0.125*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,4]=0.25*(1 - η)*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,5]=-0.25*(1 + ξ)*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + 0.125*(1 - η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,6]=0.125*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2))*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ)) + (1/2)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))^2/(1 - ν₀^2)
        K[6,7]=0.25*(1 + η)*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,8]=0.25*(1 + ξ)*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) + 0.125*(1 + η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,9]=0.125*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) - 0.125*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2))*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,10]=-0.25*(1 + η)*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,11]=0.25*(1 - ξ)*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)) - 0.125*(1 + η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[6,12]=0.125*(0.125*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) - 0.125*(-0.125*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2))*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))) + (1/2)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[7,1]=-0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,2]=-0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,3]=0.03125*(1 + η)*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.03125*(1 + η)*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[7,4]=0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 + ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,5]=0.03125*(1 + ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[7,6]=0.03125*(1 + η)*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) - 0.03125*(1 + η)*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[7,7]=0.0625*(1 + η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 + ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,8]=0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,9]=0.03125*(1 + η)*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.03125*(1 + η)*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[7,10]=-0.0625*(1 + η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,11]=0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[7,12]=0.03125*(1 + η)*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.03125*(1 + η)*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[8,1]=-0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2)
        K[8,2]=-0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2) - 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[8,3]=-0.03125*(1 + ξ)*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) + 0.03125*(1 + ξ)*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[8,4]=0.0625*(1 + ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[8,5]=-0.0625*(1 + ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[8,6]=-0.03125*(1 + ξ)*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 + η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2) + 0.03125*(1 + ξ)*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[8,7]=0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[8,8]=0.0625*(1 + ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 + η)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[8,9]=-0.03125*(1 + ξ)*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2) + 0.03125*(1 + ξ)*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[8,10]=0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[8,11]=-0.03125*(1 + η)^2*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2)
        K[8,12]=-0.03125*(1 + ξ)*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) + 0.03125*(1 + ξ)*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[9,1]=-0.25*(1 - η)*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,2]=-0.25*(1 - ξ)*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,3]=-0.125*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,4]=0.25*(1 - η)*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,5]=-0.25*(1 + ξ)*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,6]=-0.125*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))) + (1/2)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[9,7]=0.25*(1 + η)*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,8]=0.25*(1 + ξ)*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,9]=-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))^2/(1 - ν₀^2)
        K[9,10]=-0.25*(1 + η)*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2)) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,11]=0.25*(1 - ξ)*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[9,12]=-0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2))*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[10,1]=0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[10,2]=-0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[10,3]=-0.03125*(1 + η)*E₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) + 0.03125*(1 + η)*E₀*ν₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[10,4]=-0.0625*(1 - η)*(1 + η)*E₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[10,5]=0.03125*(1 - ξ)*(1 - η)*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[10,6]=-0.03125*(1 + η)*E₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2) + 0.03125*(1 + η)*E₀*ν₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[10,7]=-0.0625*(1 + η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + ξ)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[10,8]=0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2) - 0.0625*(1 + ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2)
        K[10,9]=-0.03125*(1 + η)*E₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) + 0.03125*(1 + η)*E₀*ν₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2)
        K[10,10]=0.0625*(1 + η)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - ξ)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[10,11]=-0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[10,12]=-0.03125*(1 + η)*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) + 0.03125*(1 + η)*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[11,1]=-0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,2]=-0.0625*(1 - ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,3]=-0.03125*(1 - ξ)*E₀*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) + 0.03125*(1 - ξ)*E₀*ν₀*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[11,4]=0.0625*(1 - ξ)*(1 - η)*E₀*ν₀/(1 - ν₀^2) + 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,5]=-0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2) - 0.03125*(1 - η)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,6]=-0.03125*(1 - ξ)*E₀*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 + η)*E₀*(1 - ν₀)*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2) + 0.03125*(1 - ξ)*E₀*ν₀*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)))/(1 - ν₀^2)
        K[11,7]=0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 + ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,8]=-0.03125*(1 + η)^2*E₀*(1 - ν₀)/(1 - ν₀^2) + 0.0625*(1 - ξ)*(1 + ξ)*E₀/(1 - ν₀^2)
        K[11,9]=-0.03125*(1 - ξ)*E₀*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))/(1 - ν₀^2) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))/(1 - ν₀^2) + 0.03125*(1 - ξ)*E₀*ν₀*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[11,10]=-0.0625*(1 - ξ)*(1 + η)*E₀*ν₀/(1 - ν₀^2) - 0.03125*(1 - ξ)*(1 + η)*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,11]=0.0625*(1 - ξ)^2*E₀/(1 - ν₀^2) + 0.03125*(1 + η)^2*E₀*(1 - ν₀)/(1 - ν₀^2)
        K[11,12]=-0.03125*(1 - ξ)*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2) + 0.03125*(1 - ξ)*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)
        K[12,1]=-0.25*(1 - η)*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,2]=-0.25*(1 - ξ)*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) - 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,3]=-0.125*((1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2))*((1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))) + (1/2)*E₀*(1 - ν₀)*(-0.125*((1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))) + 0.125*((1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,4]=0.25*(1 - η)*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) - 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,5]=-0.25*(1 + ξ)*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(1 - η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,6]=-0.125*(-(1 - ξ^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2))*(-(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*ξ*(1 - η)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))) + (1/2)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))*(0.125*(-(1 - ξ^2)*(0.25*(-y₁ + y₂ + y₃ - y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)) - 0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*ξ*(1 - η)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))))/(1 - ν₀^2)
        K[12,7]=0.25*(1 + η)*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) + 0.125*(1 + ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,8]=0.25*(1 + ξ)*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,9]=-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*η*(1 + ξ))*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2))*((0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - η^2) - 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) + (1/2)*E₀*(1 - ν₀)*(-0.125*((0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄))*(1 - η^2) - 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*((0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*(0.25*(-y₁ - y₂ + y₃ + y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*η*(1 + ξ)))*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,10]=-0.25*(1 + η)*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2)) + 0.125*(1 - ξ)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,11]=0.25*(1 - ξ)*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) - 0.125*(1 + η)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))/(1 - ν₀^2)
        K[12,12]=-0.125*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))*(-0.125*E₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2) + 0.125*E₀*ν₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2)) + 0.125*(0.125*E₀*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η))/(1 - ν₀^2) - 0.125*E₀*ν₀*(-(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)))/(1 - ν₀^2))*(-(1 - η^2)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄)) + 2*(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*ξ*(1 + η)) + (1/2)*E₀*(1 - ν₀)*(-0.125*(-(1 - η^2)*(-0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ - x₂ + x₃ + x₄)) + 2*(0.25*(x₁ - x₂ + x₃ - x₄) + 0.25*(-x₁ + x₂ + x₃ - x₄))*ξ*(1 + η)) + 0.125*(-(0.25*(-y₁ + y₂ + y₃ - y₄) + 0.25*(y₁ - y₂ + y₃ - y₄))*(1 - ξ^2) - 2*η*(1 - ξ)*(0.25*(-y₁ - y₂ + y₃ + y₄) - 0.25*(y₁ - y₂ + y₃ - y₄))))^2/(1 - ν₀^2)

        J[1,1]=-0.25*(1 - η)*x₁ + 0.25*(1 - η)*x₂ + 0.25*(1 + η)*x₃ - 0.25*(1 + η)*x₄
        J[1,2]=-0.25*(1 - η)*y₁ + 0.25*(1 - η)*y₂ + 0.25*(1 + η)*y₃ - 0.25*(1 + η)*y₄
        J[2,1]=-0.25*(1 - ξ)*x₁ + 0.25*(1 - ξ)*x₄ - 0.25*(1 + ξ)*x₂ + 0.25*(1 + ξ)*x₃
        J[2,2]=-0.25*(1 - ξ)*y₁ + 0.25*(1 - ξ)*y₄ - 0.25*(1 + ξ)*y₂ + 0.25*(1 + ξ)*y₃

        return K*det(J)*t
    end
    elm.Kmᵉ=sparse(hcubature(BtDB,[-1,-1],[1,1])[1])
end

function integrateKb!(elm::Quad)::SparseMatrixCSC{Float64}
    E₀,ν₀=elm.material.E,elm.material.ν
    center=elm.center
    t=elm.t
    T=elm.T[1:3,1:3]
    G=elm.material.G
    x₁,y₁,z₁=T*(elm.node1.loc-center)
    x₂,y₂,z₂=T*(elm.node2.loc-center)
    x₃,y₃,z₃=T*(elm.node3.loc-center)
    x₄,y₄,z₄=T*(elm.node4.loc-center)

    b₁=y₂-y₃
    b₂=y₃-y₄
    b₃=y₄-y₁
    b₄=y₁-y₂

    c₁=x₃-x₂
    c₂=x₄-x₃
    c₃=x₁-x₄
    c₄=x₂-x₁

    d₁=norm(elm.node3.loc-elm.node2.loc)
    d₂=norm(elm.node4.loc-elm.node3.loc)
    d₃=norm(elm.node1.loc-elm.node4.loc)
    d₄=norm(elm.node2.loc-elm.node1.loc)

    δ₁=(t/d₁)^2/(5/6*(1-ν₀)+2*(t/d₁)^2)
    δ₂=(t/d₂)^2/(5/6*(1-ν₀)+2*(t/d₂)^2)
    δ₃=(t/d₃)^2/(5/6*(1-ν₀)+2*(t/d₃)^2)
    δ₄=(t/d₄)^2/(5/6*(1-ν₀)+2*(t/d₄)^2)

    Γ=[0 0 0 -2δ₁ -c₁*δ₁ b₁*δ₁ 2δ₁ -c₁*δ₁ b₁*δ₁ 0 0 0;
       0 0 0 0 0 0 -2δ₂ -c₂*δ₂ b₂*δ₂ 2δ₂ -c₂*δ₂ b₂*δ₂;
       2δ₃ -c₃*δ₃ b₃*δ₃ 0 0 0 0 0 0 -2δ₃ -c₃*δ₃ b₃*δ₃;
      -2δ₄ -c₄*δ₄ b₄*δ₄ 2δ₄ -c₄*δ₄ b₄*δ₄ 0 0 0 0 0 0]

    Xₛ=[0 0 -b₄/(b₃*c₄-b₄*c₃) b₃/(b₃*c₄-b₄*c₃);
       b₄/(b₄*c₁-b₁*c₄) 0 0 -b₁/(b₄*c₁-b₁*c₄);
       -b₂/(b₁*c₂-b₂*c₁) b₁/(b₁*c₂-b₂*c₁) 0 0;
       0 -b₃/(b₂*c₃-b₃*c₂) b₂/(b₂*c₃-b₃*c₂) 0]

    Yₛ=[0 0 -c₄/(b₃*c₄-b₄*c₃) c₃/(b₃*c₄-b₄*c₃);
       c₄/(b₄*c₁-b₁*c₄) 0 0 -c₁/(b₄*c₁-b₁*c₄);
       -c₂/(b₁*c₂-b₂*c₁) c₁/(b₁*c₂-b₂*c₁) 0 0;
       0 -c₃/(b₂*c₃-b₃*c₂) c₂/(b₂*c₃-b₃*c₂) 0]

    α=[0 0 0 -3c₁/2/d₁^2*(1-2δ₁) 1/2/d₁^2*(b₁^2-c₁^2/2*(1-6δ₁)) 3b₁*c₁/4/d₁^2*(1-2δ₁) 3c₁/2/d₁^2*(1-2δ₁) 1/2/d₁^2*(b₁^2-c₁^2/2*(1-6δ₁)) 3b₁*c₁/4/d₁^2*(1-2δ₁) 0 0 0;
       0 0 0 0 0 0 -3c₂/2/d₂^2*(1-2δ₂) 1/2/d₂^2*(b₂^2-c₂^2/2*(1-6δ₂)) 3b₂*c₂/4/d₂^2*(1-2δ₂) 3c₂/2/d₂^2*(1-2δ₂) 1/2/d₂^2*(b₂^2-c₂^2/2*(1-6δ₂)) 3b₂*c₂/4/d₂^2*(1-2δ₂);
       3c₃/2/d₃^2*(1-2δ₃) 1/2/d₃^2*(b₃^2-c₃^2/2*(1-6δ₃)) 3b₃*c₃/4/d₃^2*(1-2δ₃) 0 0 0 0 0 0 -3c₃/2/d₃^2*(1-2δ₃) 1/2/d₃^2*(b₃^2-c₃^2/2*(1-6δ₃)) 3b₃*c₃/4/d₃^2*(1-2δ₃);
       -3c₄/2/d₄^2*(1-2δ₄) 1/2/d₄^2*(b₄^2-c₄^2/2*(1-6δ₄)) 3b₄*c₄/4/d₄^2*(1-2δ₄) 3c₄/2/d₄^2*(1-2δ₄) 1/2/d₄^2*(b₄^2-c₄^2/2*(1-6δ₄)) 3b₄*c₄/4/d₄^2*(1-2δ₄) 0 0 0 0 0 0]

    β=[0 0 0 3b₁/2/d₁^2*(1-2δ₁) 3b₁*c₁/4/d₁^2*(1-2δ₁) 1/2/d₁^2*(c₁^2-b₁^2/2*(1-6δ₁)) -3b₁/2/d₁^2*(1-2δ₁) 3b₁*c₁/4/d₁^2*(1-2δ₁) 1/2/d₁^2*(c₁^2-b₁^2/2*(1-6δ₁)) 0 0 0;
       0 0 0 0 0 0 3b₂/2/d₂^2*(1-2δ₂) 3b₂*c₂/4/d₂^2*(1-2δ₂) 1/2/d₂^2*(c₂^2-b₂^2/2*(1-6δ₂)) -3b₂/2/d₂^2*(1-2δ₂) 3b₂*c₂/4/d₂^2*(1-2δ₂) 1/2/d₂^2*(c₂^2-b₂^2/2*(1-6δ₂));
       -3b₃/2/d₃^2*(1-2δ₃) 3b₃*c₃/4/d₃^2*(1-2δ₃) 1/2/d₃^2*(c₃^2-b₃^2/2*(1-6δ₃)) 0 0 0 0 0 0 3b₃/2/d₃^2*(1-2δ₃) 3b₃*c₃/4/d₃^2*(1-2δ₃) 1/2/d₃^2*(c₃^2-b₃^2/2*(1-6δ₃));
       3b₄/2/d₄^2*(1-2δ₄) 3b₄*c₄/4/d₄^2*(1-2δ₄) 1/2/d₄^2*(c₄^2-b₄^2/2*(1-6δ₄)) -3b₄/2/d₄^2*(1-2δ₄) 3b₄*c₄/4/d₄^2*(1-2δ₄) 1/2/d₄^2*(c₄^2-b₄^2/2*(1-6δ₄)) 0 0 0 0 0 0]

    D₀=E₀*t^3/12/(1-ν₀^2)
    D=D₀*[1 ν₀ 0;
          ν₀ 1 0;
          0  0 (1-ν₀)/2]
    k=5/6
    C=k*G*t*[1 0;0 1]

    K=Matrix{Float64}(undef,12,12)
    J=Matrix{Float64}(undef,2,2)
    function BtDB(x)
        ξ,η=x[1],x[2]
        J[1,1]=-0.25*(1 - η)*x₁ + 0.25*(1 - η)*x₂ + 0.25*(1 + η)*x₃ - 0.25*(1 + η)*x₄
        J[1,2]=-0.25*(1 - η)*y₁ + 0.25*(1 - η)*y₂ + 0.25*(1 + η)*y₃ - 0.25*(1 + η)*y₄
        J[2,1]=-0.25*(1 - ξ)*x₁ + 0.25*(1 - ξ)*x₄ - 0.25*(1 + ξ)*x₂ + 0.25*(1 + ξ)*x₃
        J[2,2]=-0.25*(1 - ξ)*y₁ + 0.25*(1 - ξ)*y₄ - 0.25*(1 + ξ)*y₂ + 0.25*(1 + ξ)*y₃

        N₁⁰=1/4*(1-ξ)*(1-η)
        N₂⁰=1/4*(1+ξ)*(1-η)
        N₃⁰=1/4*(1+ξ)*(1+η)
        N₄⁰=1/4*(1-ξ)*(1+η)

        Nₛ⁰=[N₁⁰ N₂⁰ N₃⁰ N₄⁰]
        Bs=[Nₛ⁰*Xₛ*Γ;
            Nₛ⁰*Yₛ*Γ]

        dN₁dξ = -0.25*(1 - ξ)*(1 - η) + 0.25*(1 + η + ξ)*(1 - η)
        dN₁dη = -0.25*(1 - ξ)*(1 - η) + 0.25*(1 + η + ξ)*(1 - ξ)
        dN₂dξ = 0.25*(1 + ξ)*(1 - η) - 0.25*(1 + η - ξ)*(1 - η)
        dN₂dη = -0.25*(1 + ξ)*(1 - η) + 0.25*(1 + η - ξ)*(1 + ξ)
        dN₃dξ = 0.25*(1 + ξ)*(1 + η) - 0.25*(1 - η - ξ)*(1 + η)
        dN₃dη = 0.25*(1 + ξ)*(1 + η) - 0.25*(1 - η - ξ)*(1 + ξ)
        dN₄dξ = -0.25*(1 - ξ)*(1 + η) + 0.25*(1 - η + ξ)*(1 + η)
        dN₄dη = 0.25*(1 - ξ)*(1 + η) - 0.25*(1 - η + ξ)*(1 - ξ)
        dN₅dξ = 0.5*(1 - η^2)
        dN₅dη = -1.0*η*(1 + ξ)
        dN₆dξ = -1.0*ξ*(1 + η)
        dN₆dη = 0.5*(1 - ξ^2)
        dN₇dξ = -0.5*(1 - η^2)
        dN₇dη = -1.0*η*(1 - ξ)
        dN₈dξ = -1.0*ξ*(1 - η)
        dN₈dη = -0.5*(1 - ξ^2)
        # J=Matrix(1.0I,2,2)
        dN₁dx,dN₁dy=inv(J)*[dN₁dξ,dN₁dη]
        dN₂dx,dN₂dy=inv(J)*[dN₂dξ,dN₂dη]
        dN₃dx,dN₃dy=inv(J)*[dN₃dξ,dN₃dη]
        dN₄dx,dN₄dy=inv(J)*[dN₄dξ,dN₄dη]
        dN₅dx,dN₅dy=inv(J)*[dN₅dξ,dN₅dη]
        dN₆dx,dN₆dy=inv(J)*[dN₆dξ,dN₆dη]
        dN₇dx,dN₇dy=inv(J)*[dN₇dξ,dN₇dη]
        dN₈dx,dN₈dy=inv(J)*[dN₈dξ,dN₈dη]

        H₀=[0 dN₁dx     0 0 dN₂dx     0 0     dN₃dx 0 0     dN₄dx 0;
            0     0 dN₁dy 0     0 dN₂dy 0     0 dN₃dy 0     0 dN₄dy;
            0 dN₁dy dN₁dx 0 dN₂dy dN₂dx 0 dN₃dy dN₃dx 0 dN₄dy dN₄dx]

        H₁=[dN₅dx dN₆dx dN₇dx dN₈dx;
            0 0 0 0;
            dN₅dy dN₆dy dN₇dy dN₈dy]

        H₂=[0 0 0 0;
            dN₅dy dN₆dy dN₇dy dN₈dy;
            dN₅dx dN₆dx dN₇dx dN₈dx]

        Bb=-(H₀+H₁*α+H₂*β)

        Kb=transpose(Bb)*D*Bb
        Ks=transpose(Bs)*C*Bs
        K=Kb#+Ks

        return K*det(J)
    end
    Kb=sparse(hcubature(BtDB,[-1,-1],[1,1])[1])
    #left-hand system to right-hand system
    I=1:12
    J=[1,3,2,4,6,5,7,9,8,10,12,11]
    L=sparse(I,J,1.,12,12)
    for i in [1,3,4,6,7,9,10,12]
        for j in [2,5,8,11]
            Kb[i,j]=Kb[j,i]=-Kb[i,j]
        end
    end
    elm.Kbᵉ=L'*sparse(Kb)*L
end

function integrateK!(elm::Quad)::SparseMatrixCSC{Float64}
    membrane,bending=elm.membrane,elm.bending
    Kᵉ=spzeros(24,24)
    if membrane
        I=1:12
        J=[1,2,6,7,8,12,13,14,18,19,20,24]
        L=sparse(I,J,1.,12,24)
        Km=integrateKm!(elm)
        Kᵉ+=L'*Km*L
    end
    if bending
        I=1:12
        J=[3,4,5,9,10,11,15,16,17,21,22,23]
        L=sparse(I,J,1.,12,24)
        Kb=integrateKb!(elm)
        Kᵉ+=L'*Kb*L
    end
    elm.Kᵉ=Kᵉ
end

function integrateKσ(elm::Quad,σ)
    E₀,ν₀=elm.material.E,elm.material.ν
    center=elm.center
    t=elm.t
    T=elm.T[1:3,1:3]
    x₁,y₁,z₁=T*(elm.node1.loc-center)
    x₂,y₂,z₂=T*(elm.node2.loc-center)
    x₃,y₃,z₃=T*(elm.node3.loc-center)
    x₄,y₄,z₄=T*(elm.node4.loc-center)
    K=Matrix{Float64}(undef,8,8)
    J=Matrix{Float64}(undef,2,2)

    o=(node1.loc+node2.loc+node3.loc+node4.loc)/4
    n1=node1.loc+[u[1];u[2];0]
    n2=node2.loc+[u[3];u[4];0]
    n3=node3.loc+[u[5];u[6];0]
    n4=node4.loc+[u[7];u[8];0]
    pt1 = n1+n2
    pt2 = n2+n3
    csys = CSys(o, pt1, pt2)
    V=csys.T
    T̄ᵉ=zeros(8,8)
    T̄ᵉ[1:2,1:2]=V
    T̄ᵉ[3:4,3:4]=V
    T̄ᵉ[5:6,5:6]=V
    T̄ᵉ[7:8,7:8]=V


    function GtΣG(ξ,η)
        σ[1,1]= G*(-1 + (1/16)*(1 - ξ)^2*(1 - η)^2 + (1/16)*(1 - ξ)^2*(1 + η)^2 + (1/16)*(1 + ξ)^2*(1 - η)^2 + (1/16)*(1 + ξ)^2*(1 + η)^2) + (1/2)*(-1 + (1/16)*(1 - ξ)^2*(1 - η)^2 + (1/16)*(1 - ξ)^2*(1 + η)^2 + (1/16)*(1 + ξ)^2*(1 - η)^2 + (1/16)*(1 + ξ)^2*(1 + η)^2)*λ
        σ[2,2]= G*(-1 + (1/16)*(1 - ξ)^2*(1 - η)^2 + (1/16)*(1 - ξ)^2*(1 + η)^2 + (1/16)*(1 + ξ)^2*(1 - η)^2 + (1/16)*(1 + ξ)^2*(1 + η)^2) + (1/2)*(-1 + (1/16)*(1 - ξ)^2*(1 - η)^2 + (1/16)*(1 - ξ)^2*(1 + η)^2 + (1/16)*(1 + ξ)^2*(1 - η)^2 + (1/16)*(1 + ξ)^2*(1 + η)^2)*λ

        Σᵉ=[σ[1,1] 0;
            0 σ[2,2]]

        G=       [0  (-1/4)*(1 - η)              0  (-1/4)*(1 + η)               0  (1/4)*(1 - η)              0  (1/4)*(1 + η);
     (-1/4)*(1 - ξ)               0  (1/4)*(1 - ξ)               0  (-1/4)*(1 + ξ)              0  (1/4)*(1 + ξ)              0]

        J[1,1]=(-1/4)*(1 - η)*x₁ + (1/4)*(1 - η)*x₃ + (-1/4)*(1 + η)*x₂ + (1/4)*(1 + η)*x₄
        J[1,2]=(-1/4)*(1 - η)*y₂ + (1/4)*(1 - η)*y₃ + (-1/4)*(1 + η)*y₂ + (1/4)*(1 + η)*y₄
        J[2,1]=(-1/4)*(1 - ξ)*x₁ + (1/4)*(1 - ξ)*x₂ + (-1/4)*(1 + ξ)*x₃ + (1/4)*(1 + ξ)*x₄
        J[2,2]=(-1/4)*(1 + ξ)*y₃ + (1/4)*(1 + ξ)*y₄
        Gᵉ'*Σᵉ*Gᵉ*det(J)
    end
    Kσ=hquadrature(GtΣG,[-1,1],[1,1])
end

function integrateKu(elm::Quad,u)
    E₀,ν₀=elm.material.E,elm.material.ν
    center=elm.center
    t=elm.t
    T=elm.T[1:3,1:3]
    x₁,y₁,z₁=T*(elm.node1.loc-center)
    x₂,y₂,z₂=T*(elm.node2.loc-center)
    x₃,y₃,z₃=T*(elm.node3.loc-center)
    x₄,y₄,z₄=T*(elm.node4.loc-center)
    K=Matrix{Float64}(undef,8,8)
    J=Matrix{Float64}(undef,2,2)

    o=(node1.loc+node2.loc+node3.loc+node4.loc)/4
    n1=node1.loc+[u[1];u[2];0]
    n2=node2.loc+[u[3];u[4];0]
    n3=node3.loc+[u[5];u[6];0]
    n4=node4.loc+[u[7];u[8];0]
    pt1 = n1+n2
    pt2 = n2+n3
    csys = CSys(o, pt1, pt2)
    V=csys.T
    T̄ᵉ=zeros(8,8)
    T̄ᵉ[1:2,1:2]=V
    T̄ᵉ[3:4,3:4]=V
    T̄ᵉ[5:6,5:6]=V
    T̄ᵉ[7:8,7:8]=V
end

# function integrateK(elm,p_delta=false,σ=nothing,large_deform=false,u=nothing)
#     K=integrateK(elm)
#     if p_delta
#         K+=integrateKσ(elm,σ)
#     elseif large_deform
#         K+=integrateKu(elm,u)
#     end
#     return K
# end

function integrateM!(elm::Quad)
    membrane,bending=elm.membrane,elm.bending
    ρ=elm.material.ρ
    center=elm.center
    t=elm.t
    T=elm.T[1:3,1:3]
    x₁,y₁,z₁=T*(elm.node1.loc-center)
    x₂,y₂,z₂=T*(elm.node2.loc-center)
    x₃,y₃,z₃=T*(elm.node3.loc-center)
    x₄,y₄,z₄=T*(elm.node4.loc-center)
    M=Matrix{Float64}(undef,24,24)
    J=Matrix{Float64}(undef,2,2)
    function ρNtN(x)
        ξ,η=x[1],x[2]
        # include(joinpath(dirname(@__FILE__),"jacobi.jl"))
        J[1,1]=(-1/4)*(1 - η)*x₁ + (1/4)*(1 - η)*x₃ + (-1/4)*(1 + η)*x₂ + (1/4)*(1 + η)*x₄
        J[1,2]=(-1/4)*(1 - η)*y₂ + (1/4)*(1 - η)*y₃ + (-1/4)*(1 + η)*y₂ + (1/4)*(1 + η)*y₄
        J[2,1]=(-1/4)*(1 - ξ)*x₁ + (1/4)*(1 - ξ)*x₂ + (-1/4)*(1 + ξ)*x₃ + (1/4)*(1 + ξ)*x₄
        J[2,2]=(-1/4)*(1 + ξ)*y₃ + (1/4)*(1 + ξ)*y₄

        M*det(J)
    end
    elm.Mᵉ=sparse(hcubature(ρNtN,[-1,1],[1,1])[1])*t
end

function integrateP(elm,elm_force)
    N=elm.N
    B=elm.B
    D=elm.D
    J=elm.J

    f,s,σ₀,ϵ₀=elmforce.f,elmforce.s,elmforce.σ₀,elmforce.ϵ₀
    a=[-1 -1]
    b=[1 1]

    function f1(x)
        N'*f*det(J)
    end

    function f2(x)
        N'*s*det(J)
    end

    function f3(x)
        x->B'*σ₀*det(J)
    end

    function f4(x)
        x->B'*D*ϵ₀*det(J)
    end

    Pᵉ_f=hquadrature(f1,a,b)
    Pᵉ_s=hquadrature(f2,[-1],[1])#ξ=-1
    Pᵉ_σ₀=hquadrature(f3,a,b)
    Pᵉ_ϵ₀=hquadrature(f4,a,b)

    Pᵉ_f,Pᵉ_s,Pᵉ_σ₀,Pᵉ_ϵ₀=calc_force(elm,elm_force) #Pᵉ=Pᵉf+Pᵉs+Pᵉσ₀+Pᵉϵ₀
    Pᵉ=Pᵉ_f+Pᵉ_s+Pᵉ_σ₀+Pᵉ_ϵ₀
end

end
